
#include "stdio.h"
#include "ctype.h"
#include "shead.h"           /* LATTICE C only -- RGJ */
#include "afr.h"             /* Ambiguous File References */

#include "c:gf_doc.h"

#ifndef stdin
#define stdin STDIN
#define stdout STDOUT
#define stderr STDERR
#define DeSmet 1
#endif

#define LMAX	512
#define PMAX	256

#define CHAR	1
#define BOL	2
#define EOL	3
#define ANY	4
#define CLASS	5
#define NCLASS	6
#define STAR	7
#define PLUS	8
#define MINUS	9
#define ALPHA	10
#define DIGIT	11
#define NALPHA	12
#define PUNCT	13
#define RANGE	14
#define ENDPAT	15

#define BIN_MODE 0x8000;
#define TXT_MODE 0;

int	cflag;
int	fflag;
int	nflag;
int	vflag;
int     rflag;                     /* replace flag  - RGJ  */
int     mflag;                     /* matches only with "r" */
int	nfile;
int     bflag;

int	debug	=	0;	   /* Set for debug code      */

char	*pp;

#ifndef vms
char	file_name[81];
#endif

char	lbuf[LMAX];
char	rbuf[2*PMAX];   /* buffer for replacement string - RGJ */
char	pbuf[PMAX+1];

/***/

main(argc, argv)
char *argv[];
{
   register char   *p;
   register int    c, i;
   int		   gotpattern;
   int		   gotcha;
   extern int _fmode;

   FILE 	   *f;

   if (argc <= 1)
      usage("No arguments");
   if (argc == 2 && argv[1][0] == '?' && argv[1][1] == 0) {
      help(documentation);
      help(patdoc);
      return;
      }
   nfile = argc-1;
   gotpattern = 0;
   for (i=1; i < argc; ++i) {
      p = argv[i];
      if (*p == '-') {
         ++p;
         if (*p == 'r') {
            ++rflag;
            str_exp(rbuf, p+1);
         } else
         while (c = *p++) {
            switch(tolower(c)) {

               case '?': help(documentation);  break;
               case 'c': ++cflag;  break;
               case 'd': ++debug;  break;
               case 'f': ++fflag;  break;
               case 'n': ++nflag;  break;
               case 'v': ++vflag;  break;
               case 'm': ++mflag;  break;
               case 'b': ++bflag;  _fmode = BIN_MODE;  break;
               default:  usage("Unknown flag");
            }
         }
         argv[i] = 0;
         --nfile;
      } else if (!gotpattern) {
	 compile(p);
	 argv[i] = 0;
	 ++gotpattern;
	 --nfile;
      }
   }
   if (!gotpattern)
      usage("No pattern");
   if (nfile == 0)
      grep(stdin, 0);
   else {
      fflag = fflag ^ (nfile > 0);
      for (i=1; i < argc; ++i) {
		if (p = argv[i]) {
	    firstdone = 0;
	    while((f=afrfopen(p, "r")) != NULL){
	       grep(f,p);
	       fclose(f,p);
	    }
/*
					    if ((f=afrfopen(p, "r")) == NULL)
					       cant(p);
					    else {
					       grep(f, p);
					       fclose(f);
					    }
*/
		}
      }
   }
}

/***/

file(s)
char *s;
{
   printf("File %s:\n", c_dta.name);  /* alter for afr's */
}

/***/

cant(s)
char *s;
{
   fprintf(stderr, "%s: cannot open\n", s);
}


/***/

help(hp)
char **hp;  /* dns added extra '*'  */
{
   register char   **dp;

   for (dp = hp; *dp; dp++)
      printf("%s\n", *dp);
}
/***/

usage(s)
char	*s;
{
   char **dp;

   fprintf(stderr, "?GREP-Error-%s\n", s);
   for (dp = documentation; *dp; dp++)
      fprintf(stderr, "%s\n", *dp);
   exit(1);
}



/***/


compile(source)
char	   *source;   /* Pattern to compile	    */
/*
 * Compile the pattern into global pbuf[]
 */
{
   register char  *s;	      /* Source string pointer	   */
   register char  *lp;	      /* Last pattern pointer	   */
   register int   c;	      /* Current character	   */
   int		  o;	      /* Temp			   */
   char 	  *spp;       /* Save beginning of pattern */
   char 	  *cclass();  /* Compile class routine	   */

   s = source;
   if (debug)
      printf("Pattern = \"%s\"\n", s);
   pp = pbuf;
   while (c = *s++) {
      /*
       * STAR, PLUS and MINUS are special.
       */
      if (c == '*' || c == '+' || c == '-') {
	 if (pp == pbuf ||
	      (o=pp[-1]) == BOL ||
	      o == EOL ||
	      o == STAR ||
	      o == PLUS ||
	      o == MINUS)
	    badpat("Illegal occurrance op.", source, s);
	 store(ENDPAT);
	 store(ENDPAT);
	 spp = pp;		 /* Save pattern end	 */
	 while (--pp > lp)	 /* Move pattern down	 */
	    *pp = pp[-1];	 /* one byte		 */
	 *pp =	 (c == '*') ? STAR :
	    (c == '-') ? MINUS : PLUS;
	 pp = spp;		 /* Restore pattern end  */
	 continue;
      }
      /*
       * All the rest.
       */
      lp = pp;	       /* Remember start       */
      switch(c) {

          case '^': store(BOL);    break;
          case '$': store(EOL);    break;
          case '.': store(ANY);    break;
          case '[': s = cclass(source, s);   break;
          case ':': 
               if (*s) {
	            c = *s++;
	            switch(tolower(c)) {

                         case 'a': store(ALPHA);  break;
                         case 'd': store(DIGIT);  break;
                         case 'n': store(NALPHA); break;
                         case 'p': store(PUNCT);  break;
                         default:  badpat("Unknown : type", source, s);
                    }
	            break;
	       }
	 else	 badpat("No : type", source, s);

      case '\\':
	 if (*s)
	    c = *s++;

      default:
	 store(CHAR);
	 store(tolower(c));
      }
   }
   store(ENDPAT);
   store(0);		    /* Terminate string     */
   if (debug) {
      for (lp = pbuf; lp < pp;) {
	 if ((c = (*lp++ & 0377)) < ' ')
	    printf("\\%o ", c);
	 else	 printf("%c ", c);
	}
	printf("\n");
   }
}

/***/

char *
cclass(source, src)
char	   *source;   /* Pattern start -- for error msg.      */
char	   *src;      /* Class start	       */
/*
 * Compile a class (within [])
 */
{
   register char   *s;	      /* Source pointer    */
   register char   *cp;       /* Pattern start	   */
   register int    c;	      /* Current character */
   int		   o;	      /* Temp		   */

   s = src;
   o = CLASS;
   if (*s == '^') {
      ++s;
      o = NCLASS;
   }
   store(o);
   cp = pp;
   store(0);			      /* Byte count	 */
   while ((c = *s++) && c!=']') {
      if (c == '\\') {                /* Store quoted char    */
	 if ((c = *s++) == '\0')      /* Gotta get something  */
	    badpat("Class terminates badly", source, s);
	 else	 store(tolower(c));
      }
      else if (c == '-' &&
	    (pp - cp) > 1 && *s != ']' && *s != '\0') {
	 c = pp[-1];		 /* Range start     */
	 pp[-1] = RANGE;	 /* Range signal    */
	 store(c);		 /* Re-store start  */
	 c = *s++;		 /* Get end char and*/
	 store(tolower(c));	 /* Store it	    */
      }
      else {
	 store(tolower(c));	 /* Store normal char */
      }
   }
   if (c != ']')
      badpat("Unterminated class", source, s);
   if ((c = (pp - cp)) >= 256)
      badpat("Class too large", source, s);
   if (c == 0)
      badpat("Empty class", source, s);
   *cp = c;
   return(s);
}

/***/

store(op)
{
   if (pp >= &pbuf[PMAX])
      error("Pattern too complex\n");
   *pp++ = op;
}


/***/

badpat(message, source, stop)
char  *message;       /* Error message */
char  *source;	      /* Pattern start */
char  *stop;	      /* Pattern end   */
{
   register int    c;

   fprintf(stderr, "-GREP-E-%s, pattern is\"%s\"\n", message, source);
   fprintf(stderr, "-GREP-E-Stopped at byte %d, '%c'\n",
	 stop-source, stop[-1]);
   error("?GREP-E-Bad pattern\n");
}



/***/

grep(fp, fn)
FILE	   *fp;       /* File to process	    */
char	   *fn;       /* File name (for -f option)  */
/*
 * Scan the file for the pattern in pbuf[]
 */
{
   register int lno, count, m;

   lno = 0;
   count = 0;
   while (fgets(lbuf, LMAX, fp)) {
      ++lno;
      m = match();
      if ((m && !vflag) || (!m && vflag) || (!mflag && rflag)) {
	 ++count;
	 if (!cflag) {
	    if (fflag && fn) {
	       file(fn);
	       fn = 0;
	    }
	    if (nflag)
	       printf("%d\t", lno);
	    printf("%s", lbuf);
	 }
      }
   }
   if (cflag) {
      if (fflag && fn)
	 file(fn);
      printf("%d\n", count);
   }
}


/***/

match()
/*
 * Match the current line (in lbuf[]), return 1 if it does.
 */
{
   register char    *l;	      /* Line pointer	    */
   char             *pmatch();
   char             *ematch;  /* End of matched string */
   int              fmatch;

   fmatch =0;
   for (l = lbuf; *l; l++) {
     if (ematch = pmatch(l, pbuf))
          if (!rflag)
               return(1);
          else {
               repl(l, ematch, rbuf);         /* replace */
               fmatch =1;
          }
   }
   if (!fmatch) return(0);
   else         return(1);
}

repl(s,e,r)      /* replace from s to e with r */
char *s;         /* source string */
char *e;         /* end of match  */
char *r;         /* replacement string */
{
     char           dbuf[PMAX];    /* dummy buffer */

     strcpy(dbuf, e   );       /* Save tail of source */
     strcpy(s,    rbuf);      /* Append replacement  w/ null*/
     strcat(s,    dbuf);      /* Restore tail */
}



char *
pmatch(line, pattern)
char		   *line;     /* (partial) line to match      */
char		   *pattern;  /* (partial) pattern to match   */
{
   register char   *l2;	      /* Current line pointer	      */
   register char   *p;	      /* Current pattern pointer      */
   register char   c;	      /* Current character	      */
   char 	   *e;	      /* End for STAR and PLUS match  */
   int		   op;	      /* Pattern operation	      */
   int		   n;	      /* Class counter		      */
   char 	   *are;      /* Start of STAR match	      */

   l2 = line;
   if (debug > 1)
      printf("pmatch(\"%s\")\n", line);
   p = pattern;
   while ((op = *p++) != ENDPAT) {
      if (debug > 1)
	 printf("byte[%d] = 0%o, '%c', op = 0%o\n",
	       l2-line, *l2, *l2, op);
      switch(op) {

     case CHAR:     if (f_tolower(*l2++) != *p++) return(0); break;
     case BOL: if (l2 != lbuf)    return(0);  break;
     case EOL: if (*l2 != '\0')   return(0);  break;
     case ANY: if (*l2++ == '\0') return(0);  break;
     case DIGIT:    if ((c = *l2++) < '0' || (c > '9')) return(0);break;
     case ALPHA:    c = f_tolower(*l2++);
          if (c < 'a' || c > 'z') return(0); break;

     case NALPHA:   c = f_tolower(*l2++);
          if (c >= 'a' && c <= 'z')     break;
          else if (c < '0' || c > '9') return(0); break;

     case PUNCT:    c = *l2++;if (c == 0 || c > ' ') return(0);  break;

     case CLASS:
     case NCLASS: c = f_tolower(*l2++); n = *p++ & 0377;
          do {
	       if (*p == RANGE) { p += 3; n -= 2; 
                    if (c >= p[-2] && c <= p[-1]) break;
	       } else if (c == *p++) break;
	  } while (--n > 1);

          if ((op == CLASS) == (n <= 1)) return(0);
	  if (op == CLASS) p += n - 2;   break;

     case MINUS: e = pmatch(l2, p);	 /* Look for a match	*/
          while (*p++ != ENDPAT); /* Skip over pattern	*/
	  if (e) 		 /* Got a match?	*/
	    l2 = e;		 /* Yes, update string	*/
	 break; 		 /* Always succeeds	*/

      case PLUS:		 /* One or more ...	*/
	 if ((l2 = pmatch(l2, p)) == 0)
	    return(0);		 /* Gotta have a match	*/
      case STAR:		 /* Zero or more ...	*/
	 are = l2;		 /* Remember line start */
	 while (*l2 && (e = pmatch(l2, p)))
	    l2 = e;		 /* Get longest match	*/
	 while (*p++ != ENDPAT); /* Skip over pattern	*/
	 while (l2 >= are) {	 /* Try to match rest	*/
	    if (e = pmatch(l2, p))
	       return(e);
	    --l2;		 /* Nope, try earlier	*/
	 }
	 return(0);		 /* Nothing else worked */

      default:
	 printf("Bad op code %d\n", op);
	 error("Cannot happen -- match\n");
      }
   }
   return(l2);
}

/***/

error(s)
char *s;
{
   fprintf(stderr, "%s", s);
   exit(1);
}

/***/

#ifdef DeSmet
fclose(file)
FILE file;
{
     return(close(file));
}
#endif
/***/

f_tolower(c)             /* Lattice C macro replacement -- RGJ */
char c;
{
     return(isupper(c) ? ((c) + ('a' - 'A')) : (c));
}
str_exp(d, s)
char *s;   /* Source string */
char *d;   /* Destination string */
{
     char      c, *d0;
     int       n, i;

     d0 = d;
     if (debug) fprintf(stderr,"expanding %s\n", s);
     while (c = *s++)
           if (c == '\\') {
                if isdigit(c = *s) {
                     n = 0;
                     for (i = 0; isdigit(c = *s++) && i <= 2 ; i++) {
                          n = 10*n + (c - '0');
                     }
                     *d++ = (char)n;
                     s--;
                } else switch (c = *s++) {
                     case 't'  : *d++ = '\t'; break;
                     case 'n'  : *d++ = '\n'; break;
                     case 'r'  : *d++ = '\r'; break;
                     case '\\' : *d++ = '\\'; break;
                     default   : *d++ = c   ; break;
                }
           } else *d++ = c;
     *d = '\0';
     if (debug) fprintf(stderr,"expanded to %s\n", d0);
}



